@@ -6,7 +6,7 @@ from typing import Optional
 from google.cloud.storage.blob import Blob
 
 from gs_chunked_io.config import default_chunk_size, reader_retries
-from gs_chunked_io.async_collections import AsyncQueue
+from gs_chunked_io.async_collections import AsyncSet, AsyncQueue
 
 
 _BLOB_CHUNK_SIZE_UNIT = 262144
@@ -46,12 +46,14 @@ class Reader(io.IOBase):
         start_chunk = chunk_number * self.chunk_size
         end_chunk = start_chunk + self.chunk_size - 1
         if chunk_number == (self.number_of_chunks - 1):
-            expected_part_size = self.blob.size % self.chunk_size
+            expected_chunk_size = self.blob.size % self.chunk_size
+            if 0 == expected_chunk_size:
+                expected_chunk_size = self.chunk_size
         else:
-            expected_part_size = self.chunk_size
+            expected_chunk_size = self.chunk_size
         for _ in range(reader_retries):
             data = self.blob.download_as_string(start=start_chunk, end=end_chunk)
-            if expected_part_size == len(data):
+            if expected_chunk_size == len(data):
                 return data
         raise ValueError("Unexpected part size")
 
@@ -128,5 +130,10 @@ def for_each_chunk_async(blob: Blob, chunk_size: int=default_chunk_size, threads
         return chunk_number, data
 
     with ThreadPoolExecutor(max_workers=threads) as e:
-        for chunk_number, chunk in e.map(fetch_chunk, range(reader.number_of_chunks)):
-            yield chunk_number, chunk
+        chunks = AsyncSet(e, threads)
+        for chunk_number in range(reader.number_of_chunks):
+            for chunk in chunks.consume_finished():
+                yield chunk
+            chunks.put(fetch_chunk, chunk_number)
+        for chunk in chunks.consume():
+            yield chunk
